#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

typedef uint8_t u8;
typedef uint32_t u32;

static struct {
	u8 accl;
	u8 accb[255];
	FILE* file;

	int bits;
	int endCode;
	int clearCode;
	int freeCode;

	u32 accc;
	int accw;
	int cbits;
	int cmax;
} ctx;

///////////////////////////////////////////// outb

static void outb_init() { ctx.accl = 0; }

static void outb_flush()
{
	if (ctx.accl) {
		fwrite(&ctx.accl, 1, 1 + ctx.accl, ctx.file);
		ctx.accl = 0;
	}
}

static void outb(u8 b)
{
	ctx.accb[ctx.accl++] = b;
	if (ctx.accl == 255)
		outb_flush();
}

///////////////////////////////////////////// outc

enum {
	BITS = 12,
	maxcmax = 1 << BITS,
	maxbits = BITS
};

#define MAXCODE(n) ((1 << (n)) - 1)

static void outc(int code)
{
	ctx.accc |= (u32)code << ctx.accw;
	ctx.accw += ctx.cbits;
	while (ctx.accw >= 8) {
		outb(ctx.accc & 0xFF);
		ctx.accc >>= 8;
		ctx.accw -= 8;
	}
	if (code == ctx.clearCode) {
		ctx.cbits = ctx.bits;
		ctx.cmax = MAXCODE(ctx.cbits);
		return;
	}
	if (code == ctx.endCode) {
		if (ctx.accw)
			outb(ctx.accc & 0xFF);
		outb_flush();
		return;
	}
	if (ctx.freeCode <= ctx.cmax)
		return;
	ctx.cbits++;
	if (ctx.cbits == maxbits)	//??
		ctx.cmax = maxcmax;
	else
		ctx.cmax = MAXCODE(ctx.cbits);
}

static void outc_init()
{
	ctx.accc = ctx.accw = 0;
	ctx.cbits = ctx.bits;
	ctx.cmax = MAXCODE(ctx.cbits);
}

///////////////////////////////////////////// hash

#define HSIZE 8191

static long *hashTab;
static int *codeTab;

static inline int hash(long d)
{
	d = (d << 13) - d; //*=8191
	d ^= d >> 15;
	return d & 8191;
}

static int code_of(int p,int s)
{
	long c = (long)p << 8 | s;
	int i = hash(c);
	if (hashTab[i] == c)
		return codeTab[i];

	while (hashTab[i] >= 0) {
		i += p + 1;
		if (i >= HSIZE)
			i -= HSIZE;
		if (hashTab[i] == c)
			return codeTab[i];
	}

	if (ctx.freeCode >= maxcmax)
		return -1;
	codeTab[i] = ctx.freeCode;
	hashTab[i] = c;
	return 0;
}

static void hash_clear()
{
	int i;
	for (i=0; i<HSIZE; i++)
		hashTab[i] = -1;
}


///////////////////////////////////////////// compress
#include <err.h>

int compress(int init_bits, FILE* outfile, int (*readValue)())
{
	int p, s;

	ctx.bits = init_bits + 1;
	ctx.file = outfile;

	fputc(init_bits, ctx.file);

	outb_init();
	outc_init();

	ctx.clearCode = 1 << init_bits;
	ctx.endCode = ctx.clearCode + 1;
	ctx.freeCode = ctx.clearCode + 2;

//	long hashTab_arr[HSIZE];
//	hashTab = hashTab_arr;
//	int codeTab_arr[HSIZE];
//	codeTab = codeTab_arr;
	hashTab = malloc(HSIZE * sizeof *hashTab);
	if (!hashTab) err(1, 0);
	codeTab = malloc(HSIZE * sizeof *codeTab);
	if (!codeTab) err(1, 0);
	hash_clear();

//	outc(clearCode);

	p = readValue();

	for (;;) {
		s = readValue();
		if (s == EOF)
			break;

		int c = code_of(p, s);
		if (c > 0) {
			p = c;
			continue;
		}
		outc(p);
		p = s;
		if (c < 0) {
			hash_clear();
			ctx.freeCode = ctx.clearCode + 2;
			outc(ctx.clearCode);
		} else
			ctx.freeCode++;
	}

	outc(p);
	outc(ctx.endCode);

	free(codeTab);
	free(hashTab);

	fputc(0, ctx.file); // zero-length packet

	return ferror(ctx.file) ? 0 : 1;
}
